module Ast

open X64instr

open System

type ExecutableFormat    =
  (* ExecutableFormat *) | ELF
  (* ExecutableFormat *) | A_OUT

let exeFormat = ELF

type Item    = (* Abstract Syntax Tree Item *)
  (* Item *) | (* 8-bit scalar integer register  *) R8reg  of R8           
  (* Item *) | (* 16-bit scalar integer register *) R16reg of R16           
  (* Item *) | (* 32-bit scalar integer register *) R32reg of R32            
  (* Item *) | (* 64-bit scalar integer register *) R64reg of R64           
  (* Item *) | (* 128-bit SSE vector register    *) XMMreg of X           
  (* Item *) | (* 256-bit AVX vector register    *) YMMreg of Y
 
  (* Item *) | (* effective address              *) Mem of M            
  (* Item *) | (* register-indirect call/branch  *) NR64reg of NR64         
  (* Item *) | (* memory-indirect call/branch    *) NMem  of NM64         
  (* Item *) | (* floating point value           *) Float of float      
  (* Item *) | (* quoted string                  *) String of string    
  (* Item *) | (* nothing                        *) Unit                
  (* Item *) | (* type-info, eg `@function'      *) Type of string      
  (* Item *) | (* TBD name                       *) Iexpr of Iexpr 
  (* Item *) | (* TBD name                       *) Literal of Iexpr 
with override x.ToString () = 
                         match x
                          with | Unit -> "" 
                               | Iexpr(e) -> e.ToString()
                               | Type(s) -> "@" + s
                               | Mem(m) -> string(m)
                               | NR64reg(i) -> "*" + string (i.S)
                               | NMem(m) -> "*" + string (m.S)
                               | R8reg(i) -> string(i)
                               | R16reg(i) -> string(i)
                               | R32reg(i) -> string(i)
                               | R64reg(i) -> string(i)
                               | XMMreg(v) -> string(v)
                               | YMMreg(v) -> string(v)
                               | Literal(i) -> sprintf "$%s" (i.ToString())
                               | Float(f) -> sprintf "%f" f 
                               | String(s) -> let w = new System.IO.StringWriter()
                                              w.Write('\"')
                                              for i = 0 to s.Length - 1 do match s.[i] with '\b' -> w.Write("\\b")
                                                                                          | '\f' -> w.Write("\\f")
                                                                                          | '\n' -> w.Write("\\n")
                                                                                          | '\r' -> w.Write("\\r")
                                                                                          | '\t' -> w.Write("\\t")
                                                                                          | '\"' -> w.Write("\\\"")
                                                                                          | c -> if System.Char.IsControl(c) then fprintf w "\\x%x" (int c) else w.Write(c)
                                              w.Write('\"')
                                              w.ToString()

(***)

let Item_of_X(x)    = Item.XMMreg(x)
let Item_of_Y(x)    = Item.YMMreg(x)

let Item_of_R8(x)   = Item.R8reg(x)
let Item_of_R16(x)  = Item.R16reg(x)
let Item_of_R32(x)  = Item.R32reg(x)
let Item_of_R64(x)  = Item.R64reg(x)

let Item_of_R = function Rbyte(x) -> Item_of_R8(x) | Rword(x) -> Item_of_R16(x) | Rlong(x) -> Item_of_R32(x) | Rquad(x) -> Item_of_R64(x)


let Item_of_M(x)            = Item.Mem(x)

let Item_of_M8(M8.S(x))     = Item.Mem(x)
let Item_of_M16(M16.S(x))   = Item.Mem(x)
let Item_of_M32(M32.S(x))   = Item.Mem(x)
let Item_of_M64(M64.S(x))   = Item.Mem(x)
let Item_of_M128(M128.S(x)) = Item.Mem(x)
let Item_of_M256(M256.S(x)) = Item.Mem(x)


let Item_of_NR64(x)  = Item.NR64reg(x)
let Item_of_NM64(x)  = Item.NMem(x)

let Item_of_Lit(x) = Item.Literal(x.K)

let Item_of_Int64(x:int64) = Item.Literal (Iexpr.Int (bigint x))

let Item_of_String(x) = Item.Iexpr(Iexpr.Id(x))
let Item_of_Iexpr(x)  = Item.Iexpr(x)

(***)

let (|Match_Lit|_|) = function Literal(x) -> Some({Lit.K = x}) | _ -> None

let (|Match_Iexpr|_|) = function Iexpr x -> Some x | _ -> None
let (|Match_Int64|_|) = function Literal(Iexpr.Int(x)) -> try  Some(int64(x))
                                                          with _ -> None
                               | _ -> None

let (|Match_NM64|_|) = function NMem(x) -> Some(x)  | _ -> None
let (|Match_NR64|_|) = function NR64reg(x) -> Some(x) | _ -> None

let (|Match_M|_|) = function Mem(x) -> Some(x)
                           | Iexpr(x) -> Some(M.K(x))
                           | _ -> None

let (|Match_M8|_|)   = function Match_M(x) -> Some(M8.S(x)) | _ -> None
let (|Match_M16|_|)  = function Match_M(x) -> Some(M16.S(x)) | _ -> None
let (|Match_M32|_|)  = function Match_M(x) -> Some(M32.S(x)) | _ -> None
let (|Match_M64|_|)  = function Match_M(x) -> Some(M64.S(x)) | _ -> None
let (|Match_M128|_|) = function Match_M(x) -> Some(M128.S(x)) | _ -> None
let (|Match_M256|_|) = function Match_M(x) -> Some(M256.S(x)) | _ -> None

let (|Match_R8|_|)  = function R8reg(i)  -> Some(i) | _ -> None
let (|Match_R16|_|) = function R16reg(i) -> Some(i) | _ -> None
let (|Match_R32|_|) = function R32reg(i) -> Some(i) | _ -> None
let (|Match_R64|_|) = function R64reg(i) -> Some(i) | _ -> None

let (|Match_R|_|) = function R8reg(i)  -> Some(Rbyte(i))
                           | R16reg(i) -> Some(Rword(i))
                           | R32reg(i) -> Some(Rlong(i))
                           | R64reg(i) -> Some(Rquad(i))
                           | _         -> None

let (|Match_X|_|) = function XMMreg(v) -> Some(v) | _ -> None
let (|Match_Y|_|) = function YMMreg(v) -> Some(v) | _ -> None

(***)

type ELine    = Tagged of string * Item list
  (* ELine *) | Label of string
with member x.AddImpliedLiteral = match x with Tagged(name,(Literal _)::_) -> x
                                             | Tagged(name,args) -> match GetImpliedLiteral name args.Length
                                                                     with | Some(name,l) -> Tagged(name, (Literal(Iexpr.Int(bigint(l))))::args)
                                                                          | None    -> x
                                             | _ -> x

     member x.UsesIndirectAddressing = match x with Label(_) -> false
                                                  | Tagged(_,args) -> args |> List.exists (function NR64reg(_) | NMem(_) -> true | _ -> false)
                                             
(***)

let normlDirectiveDict = let d = new System.Collections.Generic.Dictionary<string,string>()
                         [("float","single");
                          ("int",  "long");
                          ("hword","short");
                          ("2byte","short");
                          ("4byte","long");
                          ("8byte","quad");
                          ("globl","global");
                          ("skip", "space");
                          ("align",match exeFormat with ELF -> "balign" | A_OUT -> "p2align")
                         ] |> List.iter d.Add
                         d

type ILine    = Instr of Instr
  (* ILine *) | ALabel of string * int (* alignment *)
  (* ILine *) | Directive of string * Item list (* string does NOT include leading point *)
with static member MakeDirective name args = let name = (name:string).ToLower()
                                             match normlDirectiveDict.TryGetValue(name) with (true,name) -> Directive(name,args)
                                                                                           | (false,_)   -> Directive(name,args)

let ILineIsDirective = function Directive _ -> true | Instr _ | ALabel _ -> false

(***)

let makeHashedTest(xs:string []) = let set = System.Collections.Generic.HashSet<string>()
                                   xs |> Array.iter (set.Add >> ignore)
                                   fun(s:string) -> set.Contains(s)

(***)

let DataDirectives = [| "byte"; "short"; "long"; "quad"; "single"; "double"; "tfloat" |]
let IsDataDirective = makeHashedTest(DataDirectives)
let (|DataDirective|_|) = function Directive(d,args) when IsDataDirective(d) -> Some(d,args)
                                 | _ -> None
                      
let (|AlignmentDirective|_|) = function Directive(("p2align"|"p2alignw"|"p2alignl"),args) -> match args with Iexpr(Iexpr.Int(i))::_ -> Some(1 <<< (int(i)))
                                                                                                           | [] -> Some(1)
                                                                                                           | _ -> failwith "Ast.AlignmentDirective: unknown p2align format"
                                      | Directive(("balign"|"balignw"|"balignl"),args) -> match args with Iexpr(Iexpr.Int(i))::_ -> Some(int(i))
                                                                                                           | [] -> Some(1)
                                                                                                           | _ -> failwith "Ast.AlignmentDirective: unknown balign format"
                                      | Directive("align",_) -> failwith "Ast.AlignmentDirective: 'align' unsupported"
                                      | _ -> None

(***)

let IsNamedSection = function "data" | "text" | "bss" -> true
                            | _ -> false

(***)

let AttributeDirectives = [| "type"; "global";
                             "dim"; "line"; "scl"; "size"; "tag"; "weak"; (* COFF *)
                             "desc"; "stabd"; "stabn"; "stabs"; (* a.out *)
                           |]
let IsAttributeDirective = makeHashedTest(AttributeDirectives)
let (|AttributeDirective|_|) = function Directive(d,args) when IsAttributeDirective(d) -> Some(d,args)
                                      | _ -> None                           

(***)

let UnsupportedDirectives = [| "altmacro"; "endr";"equ";"equiv"; "file"; "incbin";"include"; "popsection";"previous";"pushsection"; "rept"; "set";"struct";"subsection" |]
let IsUnsupportedDirective = makeHashedTest(UnsupportedDirectives)
let (|UnsupportedDirective|_|) = function Directive(d,args) when IsUnsupportedDirective(d) -> Some(d)
                                        | _ -> None

(***)

let IgnoredDirectives = [| "abort"; "eject"; "extern"; "file"; "ident"; "line"; "ln"; "loc"; "loc_mark_labels"; "mri"; "noaltmacro"; "cfi_adjust_cfa_offset"; "cfi_escape"; "cfi_lsda";
                           "cfi_def_cfa"; "cfi_def_cfa_offset"; "cfi_def_cfa_register"; "cfi_endproc";  "cfi_offset"; "cfi_personality"; "cfi_register"; "cfi_rel_offset";
                           "cfi_remember_state"; "cfi_restore"; "cfi_return_column"; "cfi_same_value"; "cfi_signal_frame"; "cfi_startproc"; "cfi_undefined"; "cfi_window_save";
                        |]
let IsIgnoredDirective = makeHashedTest(IgnoredDirectives)
let (|IgnoredDirective|_|) = function Directive(d,args) when IsIgnoredDirective(d) -> Some(d)
                                    | _ -> None
                         
(***)

type Section = string * Item list * ILine []

(***)

type Program = Section []

(***)

let asmIndent = 16
let printAsmIndent w n = let spaces = max 1 (asmIndent - n)
                         for i = 1 to spaces do fprintf w " "

let Write_ASTlineargs w = function [] -> ()
                                 | arg::args -> fprintf w "%s" (arg.ToString())
                                                args |> List.iter (fun(arg) -> fprintf w ", %s" (arg.ToString()))

let WriteELine_to (line:ELine) (w:IO.TextWriter) =
                  match line
                   with | Label(l)       -> fprintf w "%s:" l 
                        | Tagged(t,args) -> fprintf w "\t%s" t
                                            printAsmIndent w t.Length
                                            Write_ASTlineargs w args

let WriteELine0_to (line:ELine) (w:IO.TextWriter) =
                   match line
                    with | Label(l)       -> fprintf w "%s:" l 
                         | Tagged(t,args) -> fprintf w "\t%s\t" t
                                             Write_ASTlineargs w args

(* END *)
