module ListScheduling

open System
open System.Collections.Generic

open NXyn
open NXyn.Aux

open X64instr
open InstrIO
open MoreInstrIO


(* execution resource *)
type XRes =  | Fadd | Fmul | Fstore                             (* floating-point adder/multiplier/store *)
  (* XRes *) | Dcod                                             (* decoding slot *)
  (* XRes *) | Load                                             (* data load *)
  (* XRes *) | Alu0 | Alu1 | Alu2                               (* ALU pipes 0/1/2 *)
  (* XRes *) | Imul                                             (* integer multiplier *)
  (* XRes *) | P0 of Int32                                      (* execution port 0 in <N> cycles *)
  (* XRes *) | P1 of Int32                                      (* execution port 1 in <N> cycles *)
  (* XRes *) | P2 of Int32                                      (* execution port 2 in <N> cycles *)
  (* XRes *) | P3 of Int32                                      (* execution port 3 in <N> cycles *)
  (* XRes *) | P4 of Int32                                      (* execution port 4 in <N> cycles *)
  (* XRes *) | P5 of Int32                                      (* execution port 5 in <N> cycles *)

(* execution resource use *)
type XUse =  | ALL of list<XUse>                                (* use all specified resources *)
  (* XUse *) | FIRST of list<XUse>                              (* use first available resources *)
  (* XUse *) | NOT of XUse                                      (* negation as finite failure *)
  (* XUse *) | FALSE                                            (* always fail = ALL[] *)
  (* XUse *) | TRUE                                             (* always succeed = FIRST[] *)
  (* XUse *) | COND of XUse * XUse * XUse                       (* conditional = FIRST[ALL[<1>; <2>]; ALL[NOT(<1>); <3>]] *)
  (* XUse *) | PLUS of Int32                                    (* add <N> to folded integer value *)
  (* XUse *) | TAKE of XRes * Int32                             (* occupy resource <X> for <N> cycles *)

type Algorithm =  | Thruput                                     (* try to maximize instruction throughput *)
  (* Algorithm *) | CritPathCC                                  (* pick instruction with longest critical path length *)

type Parameters = { XResources           : XRes list            (* decode/issue/execute resources *)
                    MaxInstrsPerCycle    : Int32                (* maximum instruction thruput per cycle *)
                    MaxIbytesPerCycle    : Int32                (* maximum number of bytes decoded/cycle *)
                    MaxReadPorts         : Int32                (* number of read ports per cycle *)
                    ReadPortNotNeededFor : Int32                (* a read port is NOT needed for <1> cycles after the result is available *)
                    Mode                 : Algorithm            (* scheduling algorithm: Thruput | CritPathCC *)
                    MaxOoOcycles         : Int32                (* maximum # of cycles an instruction can be issued before its operands become available *)
                    GetXUse              : Instr -> XUse        (* mapping: Instruction -> executionResourceUse *)
                    GetLatency           : Instr -> Int32       (* mapping: Instruction -> latency *)
                    NextXRes             : XRes -> XRes }       (* next cycle: rename resources *)

type Dependency =  | WAR                                        (* write after read *)
  (* Dependency *) | RAW                                        (* read after write *)
  (* Dependency *) | WAW                                        (* write after write *)

(***)

type NbwDeps_EST  = Int32 * Int32                               (* (# backward dependencies, earliest starting time *)
type CPlen_EST_ix = Int32 * Int32 * Int32                       (* (critical path length, earliest starting time, instruction index) *)

type State = { nREM          : Int32                            (* # of instruction left *)
               pastCycles    : list<Int32 * list<Int32>>        (* instructions already scheduled (reversed) *)
               noRPuntil     : Map<Operand, Int32>              (* how long will no read-port be needed to read some operand? *)
               readyInstrs   : list<CPlen_EST_ix>               (* instructions whose inputs are (soon) available *)
               waitingInstrs : Map<Int32,NbwDeps_EST>           (* instructions whose inputs are not yet available *)
               tNow          : Int32                            (* how soon is now? *)
               ixsNow        : list<Int32>                      (* (reversed) list of instruction indices in this cycle *)
               nBytesNow     : Int32                            (* total size of all instructions scheduled this cycle *)
               readPortsNow  : Set<Operand>                     (* set of operands for which a read-port is needed this cycle *)
               resources     : list<XRes * Int32> }             (* execution resources and their availability *)

(***)   

let llsched (info:Parameters) (kDict:Dictionary<String,Int32>) (instrs:Instr []) =
            (* determine instruction sizes *)
            let sizes = InstrSize.sizes_of_Instrs(instrs)

            (* register read port operands *)
            let srcRPoperands = Array.create instrs.Length Set.empty
            let dstRPoperands = Array.create instrs.Length Set.empty
            let allRPoperands = ref Set.empty
    
            (* calculate forward dependencies *)
            let fwDeps    = Array.create instrs.Length []
            let lastReads = new Dictionary<Operand, list<Int32>>()
            let lastWrite = new Dictionary<Operand, Int32 * Int32>()
            let addFWdep pred (succ,(kind,latency)) = fwDeps.[pred] <- (succ,(kind,latency)) :: fwDeps.[pred]

            instrs |> Array.iteri (fun ix instr -> let (s,ds,d,lat) = (S_of_Instr kDict false instr, DS_of_Instr(instr), D_of_Instr(instr), info.GetLatency(instr))
                                                   let (s_ds, ds_d) = (Set.union s ds, Set.union ds d)

                                                   (* accumulate read port operands *)
                                                   srcRPoperands.[ix] <- s_ds |> Set.filter Operand_needs_ReadPort
                                                   dstRPoperands.[ix] <- ds_d |> Set.filter Operand_needs_ReadPort
                                                   allRPoperands := Set.union (Set.union srcRPoperands.[ix] dstRPoperands.[ix]) !allRPoperands

                                                   (* add forward dependencies on previous instructions *)
                                                   s_ds |> Set.iter (fun(r) -> DictMisc.TryGet lastWrite r |> Option.iter (fun(wIX,wLat)       -> addFWdep wIX (ix,(RAW,wLat))))
                                                   ds_d |> Set.iter (fun(w) -> DictMisc.TryGet lastWrite w |> Option.iter (fun(wIX,_)          -> addFWdep wIX (ix,(WAW,0)))
                                                                               DictMisc.TryGet lastReads w |> Option.iter (List.iter (fun(rIX) -> addFWdep rIX (ix,(WAR,0)))))

                                                   (* update `lastReads' and `lastWrite' information *)
                                                   s_ds |> Set.iter (fun(r) -> lastReads.[r] <- ix :: (match lastReads.TryGetValue(r) with (false,_) -> [] | (true,rs) -> rs))
                                                   ds_d |> Set.iter (fun(w) -> lastReads.[w] <- []
                                                                               lastWrite.[w] <- (ix,if w = Gmemory then 0 else lat)))
            let fwDeps = fwDeps |> Array.map Array.ofList

            (* sum up backward dependencies count *)
            let bwDepsCnt = Array.create instrs.Length 0
            fwDeps |> Array.iter (Array.iter (fun(succ,_) -> bwDepsCnt.[succ] <- 1 + bwDepsCnt.[succ]))

            (* sum up critical path lengths *)
            let cpLens = Array.create instrs.Length 0
            for ix = instrs.Length - 1 downto 0 do cpLens.[ix] <- Array.fold (fun mx (succ,(kind,latency)) -> max mx (latency + cpLens.[succ])) 0 fwDeps.[ix]

            (* non-deterministically select admissible combination of execution resources *)
            let rec tryAlloc t ((res,value) as s0) xuse =
                             match xuse
                              with | TRUE          -> Some(s0)
                                   | FALSE         -> None
                                   | NOT(x)        -> tryAlloc t s0 x |> Option.fold (fun _ _ -> None) (Some(s0))
                                   | PLUS(v)       -> Some(res,value + v)
                                   | ALL(xs)       -> Some(s0) |> List.foldBack (fun x -> Option.bind (fun(s) -> tryAlloc t s x)) xs
                                   | FIRST(xs)     -> List.tryPick (tryAlloc t s0) xs
                                   | TAKE(u,n)     -> res |> ListMisc.select_first (fun (v,t0) -> v = u && t >= t0) |> Option.map (fun(a,_,z) -> ((u,t+n)::(a@z), value))
                                   | COND(u,vT,vF) -> tryAlloc t s0 u |> Option.fold (fun _ s1 -> (s1,vT)) (s0,vF) ||> tryAlloc t

            (* consider 'ready' instructions issuable this cycle *)
            let getIssuable (s:State) =
                            let t = s.tNow
                            s.readyInstrs
                            |> List.filter (fun(cpl,est,ix) -> t + info.MaxOoOcycles >= est && s.nBytesNow + sizes.[ix] <= info.MaxIbytesPerCycle)
                            |> List.map (fun(cpl,est,ix) -> (cpl,est,ix,Set.union s.readPortsNow (Set.filter (fun(o) -> s.noRPuntil.[o] < t) srcRPoperands.[ix])))
                            |> List.filter (fun(cpl,est,ix,ports) -> s.ixsNow = [] || Set.count(ports) <= info.MaxReadPorts)
                            |> List.choose (fun(cpl,est,ix,ports) -> info.GetXUse(instrs.[ix])
                                                                     |> tryAlloc t (s.resources,0)
                                                                     |> Option.map (fun(res,latBias) -> (cpl,est,ix),ports,(res,latBias+if t < est then 1 + est - t else 0)))
                           
            (* adapt earliest starting time. decrement the # backward dependencies. if zero, move the instr from waiting to ready *)
            let adaptEST ix t latencyBias ((noRPuntil,ready,waiting):(Map<Operand,Int32> * list<CPlen_EST_ix> * Map<Int32,NbwDeps_EST>)) (succ,(kind,latency)) =
                         let t1 = t  + latency + (if kind = RAW then latencyBias else 0)
                         let t2 = t1 + info.ReadPortNotNeededFor
                         let (nBWdeps,est) = waiting.[succ]
                         let (nBWdeps,est) = (nBWdeps - 1, if kind = RAW then max t1 est else est)
                         let noRPuntil = dstRPoperands.[ix] |> Set.fold (fun noRPuntil operand -> Map.add operand (max t2 noRPuntil.[operand]) noRPuntil) noRPuntil
                         let ready     = if nBWdeps > 0 then ready else (cpLens.[succ],est,succ)::ready
                         let waiting   = Map.add succ (nBWdeps,est) waiting
                         (noRPuntil,ready,waiting)

            (* boost some ready instructions to reduce register pressure *)
            let maybeBoost (waiting:Map<Int32,NbwDeps_EST>) (cpl,est,ix) =
                           match instrs.[ix] // ??? don't boost stores? only sub/add...?
                            with | XsXds(_) when fwDeps.[ix] |> Array.exists (fun (succ,(kind,latency)) -> kind = WAR && (fst(waiting.[succ]) = 1)) -> (cpl+1000,est,ix)
                                 | XsM128d(_) -> (cpl+1000,est-100,ix) // when fwDeps.[ix] |> Array.exists (fun (succ,(kind,latency)) -> kind = WAR && (fst(waiting.[succ]) = 1)) -> (cpl+1000,est,ix)
                                 | _ -> (cpl,est,ix)

            (* issue single instruction this cycle *)
            let issue1 (s:State) ((cpl,est,ix),ports1,(res1,lat1)) =
                       let ready = List.filter ((<>)(cpl,est,ix)) s.readyInstrs
                       let (noRPuntil,ready,waiting) = fwDeps.[ix] |> Array.fold (adaptEST ix s.tNow lat1) (s.noRPuntil,ready,s.waitingInstrs)
                       { s with noRPuntil     = noRPuntil
                                readyInstrs   = List.map (maybeBoost waiting) ready
                                waitingInstrs = waiting
                                ixsNow        = ix :: s.ixsNow
                                nBytesNow     = s.nBytesNow + sizes.[ix]
                                readPortsNow  = ports1
                                resources     = res1 }
                       
            (* scheduling algorithms *)
            let startNewCycle (s:State) = { s with nREM         = s.nREM - s.ixsNow.Length
                                                   pastCycles   = (s.tNow,s.ixsNow) :: s.pastCycles
                                                   tNow         = s.tNow + 1
                                                   ixsNow       = []
                                                   nBytesNow    = 0
                                                   readPortsNow = Set.empty
                                                   resources    = s.resources |> List.map (fun(u,t) -> (info.NextXRes(u),t)) }
            let finishScheduling (s:State) = s |> startNewCycle |> fun(s) -> s.pastCycles

            let rec schedThruput (s:State) =
                                 let rec getMaxSucc mxOpt (rest:list<State>) =
                                                    match rest
                                                     with | [] -> mxOpt
                                                          | s::rest -> let thisValue = 100000 * s.ixsNow.Length + List.sum (List.map (fun(ix) -> cpLens.[ix]) s.ixsNow)
                                                                       if s.nREM < 1 || s.ixsNow.Length >= info.MaxInstrsPerCycle then rest
                                                                       else s |> getIssuable |> List.fold (fun nodes n -> (issue1 s n)::nodes) rest
                                                                       |> getMaxSucc (match mxOpt with Some(bestValue,_) when thisValue <= bestValue -> mxOpt | _ -> Some(thisValue,s))
                                 if s.nREM < 1 then s |> finishScheduling
                                               else (0,s) |> defaultArg (getMaxSucc None [s]) |> snd |> startNewCycle |> schedThruput
            let rec schedCPlen (s:State) =
                               if   s.nREM < 1                                        then s |> finishScheduling
                               elif s.ixsNow.Length >= info.MaxInstrsPerCycle         then s |> startNewCycle |> schedCPlen
                               else let issuable = getIssuable(s) in if issuable = [] then s |> startNewCycle |> schedCPlen
                                                                                      else issuable |> List.maxBy (fun((cpLen,_,_),_,_) -> cpLen) |> issue1 s |> schedCPlen

            (* set up initial scheduling state, invoke scheduler, process scheduling results, and return *)
            { nREM          = instrs.Length
              pastCycles    = []
              noRPuntil     = !allRPoperands       |> Set.fold (fun noRPuntil operand -> Map.add operand info.ReadPortNotNeededFor noRPuntil) Map.empty (* optimistic *)
              readyInstrs   = [0..instrs.Length-1] |> List.choose (fun(ix) -> if bwDepsCnt.[ix] > 0 then None else Some(cpLens.[ix],0,ix))
              waitingInstrs = bwDepsCnt            |> Array.fold (fun (ix,map) cnt -> (ix+1,Map.add ix (cnt,0) map)) (0,Map.empty) |> snd
              ixsNow        = []
              tNow          = 0
              nBytesNow     = 0
              readPortsNow  = Set.empty
              resources     = info.XResources |> List.map (fun(x) -> (x,0)) }
            |> match info.Mode with Thruput    -> schedThruput
                                  | CritPathCC -> schedCPlen
            |> fun(output) -> let arr = new ResizeArray<(* issueTime *)Int32 * (* instructionIndex *)Int32>()
                              output |> List.rev |> List.iter (fun(t,ixs) -> ixs |> List.rev |> List.iter (fun(instrix) -> arr.Add(t,instrix)))
                              arr.ToArray()
